import json
import unittest
import os
import tempfile
import shutil
from pathlib import Path
import json
import sys
import dif as difPy

class DifPyTestSuite(unittest.TestCase):
    """Test suite for difPy package functionality"""
    
    @classmethod
    def setUpClass(cls):
        """Set up test directories and validate their existence"""
        cls.base_dir = Path("testing/images")
        cls.test1_dir = cls.base_dir / "folder1"
        cls.test2_dir = cls.base_dir / "folder2"
        cls.test1_subfolder = cls.test1_dir / "subfolder"
        cls.test2_subfolder = cls.test2_dir / "subfolder"
        
        # Verify all test directories exist
        for dir_path in [cls.base_dir, cls.test1_dir, cls.test2_dir, 
                        cls.test1_subfolder, cls.test2_subfolder]:
            if not dir_path.exists():
                raise ValueError(f"Required test directory missing: {dir_path}")

    def test_base_folder_recursive(self):
        """Test searching in base folder recursively (includes all subfolders)"""
        dif = difPy.build(str(self.base_dir), recursive=True, show_progress=False)
        
        # Verify build statistics
        self.assertEqual(dif.stats["total_files"], 50, "Total file count mismatch")
        self.assertEqual(dif.stats["invalid_files"]["count"], 8, "Invalid file count mismatch")
        
        # Test duplicate search
        search_dup = difPy.search(dif, similarity="duplicates", show_progress=False)
        self.assertEqual(search_dup.stats["process"]["search"]["matches_found"]["duplicates"], 22)
        self.assertEqual(search_dup.stats["process"]["search"]["matches_found"]["similar"], 0)
        
        # Test similar search
        search_sim = difPy.search(dif, similarity="similar", show_progress=False)
        self.assertEqual(search_sim.stats["process"]["search"]["matches_found"]["duplicates"], 18)
        self.assertEqual(search_sim.stats["process"]["search"]["matches_found"]["similar"], 8)

    def test_base_folder_non_recursive(self):
        """Test searching in base folder non-recursively (excludes subfolders)"""
        dif = difPy.build(str(self.base_dir), recursive=False, show_progress=False)
        
        # Should only find files in the root directory
        self.assertEqual(dif.stats["total_files"], 0, "Total file count mismatch")
        self.assertEqual(dif.stats["invalid_files"]["count"], 0, "Invalid file count mismatch")
        
        search = difPy.search(dif, show_progress=False)
        self.assertEqual(search.stats["process"]["search"]["matches_found"]["duplicates"], 0)

    def test_separate_folders_union(self):
        """Test searching across multiple folders as union"""
        dif = difPy.build(str(self.test1_dir), str(self.test2_dir), 
                         recursive=True, in_folder=False, show_progress=False)
        
        self.assertEqual(dif.stats["total_files"], 50, "Total file count mismatch")
        self.assertEqual(dif.stats["invalid_files"]["count"], 8, "Invalid file count mismatch")
        
        # Test duplicate search
        search_dup = difPy.search(dif, similarity="duplicates", show_progress=False)
        self.assertEqual(search_dup.stats["process"]["search"]["matches_found"]["duplicates"], 22)
        
        # Test similar search
        search_sim = difPy.search(dif, similarity="similar", show_progress=False)
        self.assertEqual(search_sim.stats["process"]["search"]["matches_found"]["duplicates"], 18)
        self.assertEqual(search_sim.stats["process"]["search"]["matches_found"]["similar"], 8)

    def test_separate_folders_in_folder(self):
        """Test searching in multiple folders separately"""
        dif = difPy.build(str(self.test1_dir), str(self.test2_dir), 
                         recursive=True, in_folder=True, show_progress=False)
        
        self.assertEqual(dif.stats["total_files"], 50, "Total file count mismatch")
        self.assertEqual(dif.stats["invalid_files"]["count"], 8, "Invalid file count mismatch")
        
        # Test duplicate search - should only find duplicates within each folder
        search_dup = difPy.search(dif, similarity="duplicates", show_progress=False)
        self.assertEqual(search_dup.stats["process"]["search"]["matches_found"]["duplicates"], 2)
        
        # Test similar search
        search_sim = difPy.search(dif, similarity="similar", show_progress=False)
        self.assertEqual(search_sim.stats["process"]["search"]["matches_found"]["duplicates"], 2)
        self.assertEqual(search_sim.stats["process"]["search"]["matches_found"]["similar"], 8)

    def test_build_parameters(self):
        """Test all optional parameters for the build function"""
        # Test recursive parameter
        dif = difPy.build(str(self.base_dir), recursive=True, show_progress=False)
        self.assertTrue(dif.stats["process"]["build"]["parameters"]["recursive"])
        self.assertEqual(dif.stats["total_files"], 50)

        dif = difPy.build(str(self.base_dir), recursive=False, show_progress=False)
        self.assertFalse(dif.stats["process"]["build"]["parameters"]["recursive"])
        self.assertEqual(dif.stats["total_files"], 0)  # Only root directory files

        # Test in_folder parameter
        dif = difPy.build(str(self.test1_dir), str(self.test2_dir), 
                         in_folder=True, show_progress=False)
        self.assertTrue(dif.stats["process"]["build"]["parameters"]["in_folder"])

        dif = difPy.build(str(self.test1_dir), str(self.test2_dir), 
                         in_folder=False, show_progress=False)
        self.assertFalse(dif.stats["process"]["build"]["parameters"]["in_folder"])

        # Test limit_extensions parameter
        dif = difPy.build(str(self.base_dir), limit_extensions=True, show_progress=False)
        self.assertTrue(dif.stats["process"]["build"]["parameters"]["limit_extensions"])
        self.assertGreater(dif.stats["invalid_files"]["count"], 0)

        dif = difPy.build(str(self.base_dir), limit_extensions=False, show_progress=False)
        self.assertFalse(dif.stats["process"]["build"]["parameters"]["limit_extensions"])

        # Test px_size parameter
        dif = difPy.build(str(self.base_dir), px_size=100, show_progress=False)
        self.assertEqual(dif.stats["process"]["build"]["parameters"]["px_size"], 100)

        # Test processes parameter
        dif = difPy.build(str(self.base_dir), processes=2, show_progress=False)
        self.assertEqual(dif.stats["process"]["build"]["parameters"]["processes"], 2)

    def test_search_parameters(self):
        """Test all optional parameters for the search function"""
        dif = difPy.build(str(self.base_dir), show_progress=False)

        # Test similarity parameter variations
        search = difPy.search(dif, similarity="duplicates", show_progress=False)
        self.assertEqual(search.stats["process"]["search"]["parameters"]["similarity_mse"], 0)

        search = difPy.search(dif, similarity="similar", show_progress=False)
        self.assertEqual(search.stats["process"]["search"]["parameters"]["similarity_mse"], 5)

        search = difPy.search(dif, similarity=10.5, show_progress=False)
        self.assertEqual(search.stats["process"]["search"]["parameters"]["similarity_mse"], 10.5)

        # Test rotate parameter
        search = difPy.search(dif, rotate=True, show_progress=False)
        self.assertTrue(search.stats["process"]["search"]["parameters"]["rotate"])

        search = difPy.search(dif, rotate=False, show_progress=False)
        self.assertFalse(search.stats["process"]["search"]["parameters"]["rotate"])

        # Test same_dim parameter
        search = difPy.search(dif, same_dim=True, show_progress=False)
        self.assertTrue(search.stats["process"]["search"]["parameters"]["same_dim"])

        search = difPy.search(dif, same_dim=False, show_progress=False)
        self.assertFalse(search.stats["process"]["search"]["parameters"]["same_dim"])

        # Test processes parameter
        search = difPy.search(dif, processes=2, show_progress=False)
        self.assertEqual(search.stats["process"]["search"]["parameters"]["processes"], 2)

        # Test chunksize parameter
        search = difPy.search(dif, chunksize=50, show_progress=False)
        self.assertEqual(search.stats["process"]["search"]["parameters"]["chunksize"], 50)

    def test_parameter_error_handling(self):
        """Test error handling for invalid parameter values"""
        # Test invalid build parameters
        with self.assertRaises(Exception):
            difPy.build(str(self.base_dir), px_size=1)  # Too small
        
        with self.assertRaises(Exception):
            difPy.build(str(self.base_dir), px_size=6000)  # Too large

        with self.assertRaises(Exception):
            difPy.build(str(self.base_dir), processes=0)  # Invalid process count

        # Test invalid search parameters
        dif = difPy.build(str(self.base_dir), show_progress=False)
        
        with self.assertRaises(Exception):
            difPy.search(dif, similarity=-1)  # Negative similarity

        with self.assertRaises(Exception):
            difPy.search(dif, processes=0)  # Invalid process count

        with self.assertRaises(Exception):
            difPy.search(dif, chunksize=0)  # Invalid chunksize

    def test_invalid_files(self):
        """Test handling of invalid files"""
        dif = difPy.build(str(self.base_dir), show_progress=False)
        
        # Verify specific invalid files are detected
        invalid_files = dif.stats["invalid_files"]["logs"]
        self.assertEqual(len(invalid_files), 8)
        
        # Check for expected invalid file extensions
        invalid_extensions = [Path(f).suffix for f in invalid_files.keys()]
        expected_extensions = ['.pdf', '.svg', '.bmp', '.wmf']
        for ext in expected_extensions:
            self.assertIn(ext, invalid_extensions)

class DifPyCliTestSuite(unittest.TestCase):
    """Test suite for difPy CLI functionality"""
    
    @classmethod
    def setUpClass(cls):
        """Set up test directories and validate their existence"""
        cls.base_dir = Path("testing/images")
        cls.test1_dir = cls.base_dir / "folder1"
        cls.test2_dir = cls.base_dir / "folder2"
        cls.test1_subfolder = cls.test1_dir / "subfolder"
        cls.test2_subfolder = cls.test2_dir / "subfolder"
        cls.temp_dir = Path("testing/temp")
        
        # Verify all test directories exist
        for dir_path in [cls.base_dir, cls.test1_dir, cls.test2_dir, 
                        cls.test1_subfolder, cls.test2_subfolder]:
            if not dir_path.exists():
                raise ValueError(f"Required test directory missing: {dir_path}")
        
        # Create temp directory if it doesn't exist
        cls.temp_dir.mkdir(parents=True, exist_ok=True)

    @classmethod
    def tearDownClass(cls):
        """Clean up any remaining files in temp directory"""
        if cls.temp_dir.exists():
            for file in cls.temp_dir.glob('*'):
                if file.is_file():
                    file.unlink()
                elif file.is_dir():
                    shutil.rmtree(file)

    def setUp(self):
        """Ensure temp directory is clean before each test"""
        for file in self.temp_dir.glob('*'):
            if file.is_file():
                file.unlink()
            elif file.is_dir():
                shutil.rmtree(file)

    def test_cli_basic_functionality(self):
        """Test basic CLI functionality with default parameters"""
        # Run difPy with basic parameters
        cmd = f'python dif.py -D "{str(self.base_dir)}" -Z "{str(self.temp_dir)}"'
        exit_code = os.system(cmd)
        
        self.assertEqual(exit_code, 0, "CLI command failed to execute")
        
        # Verify output files were created
        files = list(self.temp_dir.glob('difPy_*'))
        self.assertEqual(len(files), 3, "Expected 3 output files")

    def test_cli_all_parameters(self):
        """Test CLI with all available parameters"""
        move_dir = self.temp_dir / "moved_files"
        
        # Test comprehensive parameter set
        cmd = (
            f'python dif.py '
            f'-D "{str(self.test1_dir)}" "{str(self.test2_dir)}" '
            f'-Z "{str(self.temp_dir)}" '
            f'-r True '
            f'-i True '
            f'-le True '
            f'-px 100 '
            f'-s similar '
            f'-ro True '
            f'-dim True '
            f'-mv "{str(move_dir)}" '
            f'-p False '
            f'-proc 2 '
            f'-ch 50'
        )
        
        exit_code = os.system(cmd)
        self.assertEqual(exit_code, 0, "CLI command failed to execute")
        
        # Verify output files
        files = list(self.temp_dir.glob('difPy_*'))
        self.assertEqual(len(files), 3, "Expected 3 output files")
        
        # Verify moved files directory was created
        self.assertTrue(move_dir.exists(), "Move directory not created")

    def test_cli_delete_functionality(self):
        """Test CLI delete functionality with confirmation bypass"""
        # Create test directory with duplicate files
        test_files_dir = self.temp_dir / "test_files"
        test_files_dir.mkdir(exist_ok=True)
        
        # Copy some test files to temporary location
        for file in (self.test1_dir).glob('*.jpg'):
            shutil.copy2(file, f"{test_files_dir}/1.jpg")
            shutil.copy2(file, f"{test_files_dir}/2.jpg")
            break

        # Run difPy with delete parameter
        cmd = (
            f'python dif.py '
            f'-D "{str(test_files_dir)}" '
            f'-Z "{str(self.temp_dir)}" '
            f'-d True '
            f'-sd True'  # Silent delete to bypass confirmation
        )
        
        exit_code = os.system(cmd)
        self.assertEqual(exit_code, 0, "CLI command failed to execute")
        
        # Verify results
        original_file_count = len(list(test_files_dir.glob('*')))
        self.assertLess(original_file_count, len(list(self.test1_dir.glob('*.jpg'))), 
                       "Some files should have been deleted")

    def test_cli_invalid_parameters(self):
        """Test CLI behavior with invalid parameters"""
        # Test invalid directory
        cmd = f'python dif.py -D "nonexistent_directory" -Z "{str(self.temp_dir)}"'
        exit_code = os.system(cmd)
        self.assertNotEqual(exit_code, 0, "Command should fail with invalid directory")
        
        # Test invalid pixel size
        cmd = f'python dif.py -D "{str(self.base_dir)}" -Z "{str(self.temp_dir)}" -px 0'
        exit_code = os.system(cmd)
        self.assertNotEqual(exit_code, 0, "Command should fail with invalid pixel size")
        
        # Test mutually exclusive parameters (move_to and delete)
        cmd = (
            f'python dif.py '
            f'-D "{str(self.base_dir)}" '
            f'-Z "{str(self.temp_dir)}" '
            f'-mv "{str(self.temp_dir)}/moved" '
            f'-d True'
        )
        exit_code = os.system(cmd)
        self.assertNotEqual(exit_code, 0, "Command should fail with mutually exclusive parameters")

    def test_cli_output_verification(self):
        """Test CLI output file contents"""
        # Run difPy
        cmd = f'python dif.py -D "{str(self.base_dir)}" -Z "{str(self.temp_dir)}"'
        exit_code = os.system(cmd)
        self.assertEqual(exit_code, 0, "CLI command failed to execute")
        
        # Get the most recent output files
        result_file = max(self.temp_dir.glob('difPy_*_results.json'), key=os.path.getctime)
        stats_file = max(self.temp_dir.glob('difPy_*_stats.json'), key=os.path.getctime)
        lq_file = max(self.temp_dir.glob('difPy_*_lower_quality.txt'), key=os.path.getctime)
        
        # Verify JSON structure in results file
        with open(result_file, 'r') as f:
            results = json.load(f)
            self.assertIsInstance(results, dict, "Results file should contain a dictionary")
        
        # Verify JSON structure in stats file
        with open(stats_file, 'r') as f:
            stats = json.load(f)
            self.assertIn('directory', stats, "Stats should contain 'directory' key")
            self.assertIn('total_files', stats, "Stats should contain 'total_files' key")
            self.assertIn('process', stats, "Stats should contain 'process' key")
        
        # Verify lower quality file contains content
        with open(lq_file, 'r') as f:
            content = f.read()
            self.assertGreater(len(content), 0, "Lower quality file should not be empty")

def run_tests():
    """Run the test suite"""
    # Create test suites
    suite1 = unittest.TestLoader().loadTestsFromTestCase(DifPyTestSuite)
    suite2 = unittest.TestLoader().loadTestsFromTestCase(DifPyCliTestSuite)
    
    # Combine test suites
    all_tests = unittest.TestSuite([suite1, suite2])
    
    # Run tests with detailed output
    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(all_tests)
    
    return 0 if result.wasSuccessful() else 1

if __name__ == '__main__':
    sys.exit(run_tests())